import java.io.*;
import java.util.*;


public class AVLTestTree extends AVLTree {


  
 /**
   * Checks if the tree is balanced. 
   * Will be used by autograder.
   *
   * WARNING: DO NOT MODIFY THIS METHOD!!!
   * @returns a Check construct, where bool is true if the 
   * tree is balanced, height is the tree height
   * and node is the first point of imbalance detected
   *
   */

   public  boolean isHBalanced()
   {
	   CheckStruct c= isHBalancedHelper(root); 
	   if (!c.bool)
	   {
		   System.out.println("Error: The following subtree is not height balanced or its root does not have a correct balance factor:");
		printHelper(c.node,"");
	   }
	   return c.bool;
   }		   

  /**
   * Checks if the tree is ballanced. 
   * Will be used by autograder.
   * WARNING: DO NOT MODIFY THIS METHOD!!!
   * 
   * @param root The head of the tree to be checked 
   * @returns a Check construct, where bool is true if the 
   * tree is balanced, height is the tree height
   * and node is the first point of imbalance detected
   *
   */
   private CheckStruct isHBalancedHelper(Node root)
   {
	// If null, the tree is balanced and the height is 0  
	   if (root == null) {
	      CheckStruct c = new CheckStruct();
              c.bool = true;	      
              c.height = 0;
	      return c;
	   }   
	   
	// Check if subtrees are balanced and obtain their heights   
	   CheckStruct c_left = isHBalancedHelper(root.left);
	   CheckStruct c_right = isHBalancedHelper(root.right);
	   
	// The root's height is 1 + the maximum hight of the
	// children.    
	   int rootheight = (c_left.height>=c_right.height) ?
		                     c_left.height + 1: c_right.height + 1;

	// If the left tree was unbalanced, update the height and return
	// relevant information   
	   if (c_left.bool == false) {
	     c_left.height =  rootheight;
	     return c_left;
	   }
	   
	// If the right tree was unbalanced, update the height and return
	// relevant information   
	   if (c_right.bool == false) {
	     c_right.height =  rootheight;
	     return c_right;
	   }

	// If both children are balanced, check if root is balanced  or the
	// balance factor at the root is wrong 
	   int balance = c_right.height - c_left.height;
	   if ( (balance < -1 ) || (balance > +1) || root.bf !=balance) {
	// If unbalanced, update height and mark root as the
	// first unbalanced node detected	   
	      c_left.height =  rootheight;
	      c_left.bool = false;
	      c_left.node = root; 
	      return c_left;		  
	   } else {
	// The tree under root is balanced factor is fine,
	// update height	   
              c_left.height =  rootheight;
	      c_left.bool = true;
	      c_left.node = null;
	      return c_left;		  
	   }
	// Either way c_left is sent up and c_right is left 
	// for the garbage collector   
   }

  /**
   * Checks if the tree is a BST. 
   * Will be used by autograder.
   *
   * WARNING: DO NOT MODIFY THIS METHOD!!!
   * @returns a Check structure, where bool is true if the 
   * tree is a BST, max is the maximum key, min is the minumum
   * ket and node is the root of the first non BST subtree detected
   *
   */
  

   public boolean isBST()
   {
	   CheckStruct c= isBSTHelper(root);
	   if (!c.bool)
	   {
		   System.out.println("Error: The following subtree is not a BST:");
		printHelper(c.node,"");
	   }
	   return c.bool;

   }
	   
	   
   private CheckStruct isBSTHelper(Node root)
   {

    // This should not be reached by recursion. It only occurs if the
    // method is called with a null pointer as argument. c.max and 
    // c.min are then bogus.
      if (root == null) {
        CheckStruct c = new CheckStruct();
	c.bool = true;
        return c;
      }
 
      CheckStruct c_left,c_right;

      // If there is no left subtree, then the left subtree is a BST 
      // and set max and min to be the parent value. Else call method 
      // recursively on left child.
      if (root.left == null) {
        c_left = new CheckStruct();
	c_left.bool = true;
	c_left.max = root.key;
	c_left.min = root.key;
      } else c_left = isBSTHelper(root.left);

      // If there is no right subtree, then the right subtree is a BST 
      // and set max and min to be the parent value. Else call method 
      // recursively on right child.
      if (root.right == null) {
	c_right = new CheckStruct();
	c_right.bool = true;
	c_right.max = root.key;
	c_right.min = root.key;
      } else c_right = isBSTHelper(root.right);

      int max,min;

      // Compute glogal maximum and glogal minimum
      max = (c_left.max > root.key) ? c_left.max : root.key;
      max =  (max > c_right.max) ?  max : c_right.max;

      min = (c_left.min < root.key) ? c_left.min : root.key;
      min = (min < c_right.min) ? min : c_right.min; 
     
     // If either of the subtrees are not BST, neither is root. Update
     // maximum and minimum and maximum and propagate information upwards 
      if ((c_left.bool == false) || (c_right.bool == false) ) {
        c_left.max = max;
        c_left.min = min;
	c_left.node = (c_left.bool == false) ? c_left.node : c_right.node;
        c_left.bool = false;
	return c_left;
      }  
     
      // If the root key is smaller than the minimum of the right subtree
      // or greater than the maximum of the left subtree this is not a BST
      // Update max and min and set root as the first non BST tree detected
      if ((root.key>c_right.min)|| (root.key<c_left.max)) {
        c_left.bool = false;
	c_left.max = max;
	c_left.min = min;
	c_left.node = root;
	return c_left;
      }	
	
      // O.w. all is well, update and push upwards 
      c_left.bool = true;
      c_left.max = max;
      c_left.min = min;
      c_left.node = null;
      return c_left;	      

      //c_right is left for the garbage collector
   }

   private static void printHelper(Node n, String indent) {
    
   
    if (n == null) {
      System.out.println(indent+"x");
      return;
    }
   
    String blanks="";
    for (int i=0;i<n.printNode().length();++i) blanks=blanks+" ";
    String barandblanks="|";
    for (int i=0;i<n.printNode().length()-1;++i) barandblanks = barandblanks+" ";

    if (n.parent == null) //root
    {
      printHelper(n.right, indent+blanks);
      System.out.println(indent+"*"+n.printNode());
      printHelper(n.left, indent+blanks);
      return;
    } 
    
    if (n.parent.left == n)
    {
      printHelper(n.right, indent+barandblanks);
      System.out.println(indent+"|"+n.printNode());
      printHelper(n.left, indent+blanks);
    } else
    {
      printHelper(n.right, indent+blanks);
      System.out.println(indent+"|"+n.printNode());
      printHelper(n.left, indent+barandblanks);
    }
  }//printHelper
   
   private static void simplePrint(Node n) {
    
   
    if (n == null) {
      System.out.println("x");
      return;
    }
   
    System.out.print(n.printNode()+" ");
    simplePrint(n.left);
    simplePrint(n.right);
    return;
    
  }//printHelper




  public void BSTinsert(int key, int data, int bf)
  {
    Node z = new Node(key, data);
    z.bf = bf;
    BSTInsert(z);
  } //insert
  
  /**
   * Inserts a node into the tree as BST.
   *
   * @param z The node to insert.
   */
  protected void BSTInsert(Node z)
  {
    Node y = null;
    Node x = root;
    
    while (x != null) {
      y = x;
      if (z.key <= x.key)
        x = x.left;
      else
        x = x.right;
    } //while
    
    z.parent = y;
    if (y == null)
      root = z;  // the tree had been empty
    else {
      if (z.key <= y.key)
        y.left = z;
      else
        y.right = z;
    }//else
   }	

  public int search(int k)
  {
    Node node = search(root,k);
    return node.data;
  } //search

};
